@@ -29,6 +29,7 @@ import collections
 import fnmatch
 import functools
 import logging
+import os
 import posixpath
 import urllib.parse
 from wsgiref.util import request_uri
@@ -293,26 +294,33 @@ class Property(object):
     live = None
 
     def supported_on(self, resource):
-        return (self.resource_type is None or
-                self.resource_type in resource.resource_types)
+        if self.resource_type is None:
+            return True
+        if isinstance(self.resource_type, tuple):
+            return any(rs in resource.resource_types
+                       for rs in self.resource_type)
+        if self.resource_type in resource.resource_types:
+            return True
+        return False
 
-    def is_set(self, href, resource):
+    def is_set(self, href, resource, environ):
         """Check if this property is set on a resource."""
         if not self.supported_on(resource):
             return False
         try:
-            self.get_value('/', resource, ET.Element(self.name))
+            self.get_value('/', resource, ET.Element(self.name), environ)
         except KeyError:
             return False
         else:
             return True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         """Get property with specified name.
 
         :param href: Resource href
         :param resource: Resource for which to retrieve the property
         :param el: Element to populate
+        :param environ: WSGI environment dict
         :raise KeyError: if this property is not present
         """
         raise KeyError(self.name)
@@ -338,7 +346,7 @@ class ResourceTypeProperty(Property):
 
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         for rt in resource.resource_types:
             ET.SubElement(el, rt)
 
@@ -355,7 +363,7 @@ class DisplayNameProperty(Property):
     name = '{DAV:}displayname'
     resource_type = None
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = resource.get_displayname()
 
     def set_value(self, href, resource, el):
@@ -372,10 +380,13 @@ class GetETagProperty(Property):
     resource_type = None
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = resource.get_etag()
 
 
+ADD_MEMBER_FEATURE = 'add-member'
+
+
 class AddMemberProperty(Property):
     """Provides {DAV:}add-member.
 
@@ -386,7 +397,7 @@ class AddMemberProperty(Property):
     resource_type = COLLECTION_RESOURCE_TYPE
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         # Support POST against collection URL
         el.append(create_href('.', href))
 
@@ -402,7 +413,7 @@ class GetLastModifiedProperty(Property):
     live = True
     in_allprops = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         # Use rfc1123 date (section 3.3.1 of RFC2616)
         el.text = resource.get_last_modified().strftime(
             '%a, %d %b %Y %H:%M:%S GMT')
@@ -430,7 +441,7 @@ class CreationDateProperty(Property):
     resource_type = None
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = format_datetime(resource.get_creationdate())
 
 
@@ -443,7 +454,7 @@ class GetContentLanguageProperty(Property):
     name = '{DAV:}getcontentlanguage'
     resource_type = None
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = ', '.join(resource.get_content_language())
 
 
@@ -456,7 +467,7 @@ class GetContentLengthProperty(Property):
     name = '{DAV:}getcontentlength'
     resource_type = None
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_content_length())
 
 
@@ -469,7 +480,7 @@ class GetContentTypeProperty(Property):
     name = '{DAV:}getcontenttype'
     resource_type = None
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = resource.get_content_type()
 
 
@@ -487,9 +498,9 @@ class CurrentUserPrincipalProperty(Property):
     def __init__(self, current_user_principal):
         super(CurrentUserPrincipalProperty, self).__init__()
         self.current_user_principal = ensure_trailing_slash(
-            current_user_principal)
+            current_user_principal.lstrip('/'))
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         """Get property with specified name.
 
         :param name: A property name.
@@ -497,9 +508,8 @@ class CurrentUserPrincipalProperty(Property):
         if self.current_user_principal is None:
             ET.SubElement(el, '{DAV:}unauthenticated')
         else:
-            # TODO(jelmer): Ideally this should receive
-            # SCRIPT_NAME and prefix the returned URL with that.
-            el.append(create_href(self.current_user_principal))
+            el.append(create_href(
+                self.current_user_principal, environ['SCRIPT_NAME']))
 
 
 class PrincipalURLProperty(Property):
@@ -509,7 +519,7 @@ class PrincipalURLProperty(Property):
     in_allprops = True
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         """Get property with specified name.
 
         :param name: A property name.
@@ -528,7 +538,7 @@ class SupportedReportSetProperty(Property):
     def __init__(self, reporters):
         self._reporters = reporters
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         for name, reporter in self._reporters.items():
             if reporter.supported_on(resource):
                 bel = ET.SubElement(el, '{DAV:}supported-report')
@@ -545,7 +555,7 @@ class GetCTagProperty(Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = resource.get_ctag()
 
 
@@ -565,6 +575,23 @@ class AppleGetCTagProperty(GetCTagProperty):
     name = '{http://calendarserver.org/ns/}getctag'
 
 
+class RefreshRateProperty(Property):
+    """refreshrate property.
+
+    (no public documentation, but contains an ical-style frequency indicator)
+    """
+
+    name = '{http://calendarserver.org/ns/}refreshrate'
+    resource_type = COLLECTION_RESOURCE_TYPE
+    in_allprops = False
+
+    def get_value(self, href, resource, el, environ):
+        el.text = resource.get_refreshrate()
+
+    def set_value(self, href, resource, el):
+        resource.set_refreshrate(el.text)
+
+
 LOCK_SCOPE_EXCLUSIVE = '{DAV:}exclusive'
 LOCK_SCOPE_SHARED = '{DAV:}shared'
 LOCK_TYPE_WRITE = '{DAV:}write'
@@ -815,6 +842,21 @@ class Collection(Resource):
         """
         raise NotImplementedError(self.destroy)
 
+    def set_refreshrate(self, value):
+        """Set the recommended refresh rate for this collection.
+
+        :param value: Refresh rate (None to remove)
+        """
+        raise NotImplementedError(self.set_refreshrate)
+
+    def get_refreshrate(self):
+        """Get the recommended refresh rate.
+
+        :return: Recommended refresh rate
+        :raise KeyError: if there is no refresh rate set
+        """
+        raise NotImplementedError(self.get_refreshrate)
+
 
 class Principal(Resource):
     """Resource for a DAV Principal."""
@@ -855,26 +897,34 @@ class Principal(Resource):
         """
         raise NotImplementedError(self.get_calendar_proxy_write_for)
 
+    def get_schedule_inbox_url(self):
+        raise NotImplementedError(self.get_schedule_inbox_url)
 
-def get_property_from_name(href, resource, properties, name):
+    def get_schedule_outbox_url(self):
+        raise NotImplementedError(self.get_schedule_outbox_url)
+
+
+def get_property_from_name(href, resource, properties, name, environ):
     """Get a single property on a resource.
 
     :param href: Resource href
     :param resource: Resource object
     :param properties: Dictionary of properties
+    :param environ: WSGI environ dict
     :param name: name of property to resolve
     :return: PropStatus items
     """
     return get_property_from_element(
-        href, resource, properties, ET.Element(name))
+        href, resource, properties, environ, ET.Element(name))
 
 
-def get_property_from_element(href, resource, properties, requested):
+def get_property_from_element(href, resource, properties, environ, requested):
     """Get a single property on a resource.
 
     :param href: Resource href
     :param resource: Resource object
     :param properties: Dictionary of properties
+    :param environ: WSGI environ dict
     :param requested: Requested element
     :return: PropStatus items
     """
@@ -894,9 +944,9 @@ def get_property_from_element(href, resource, properties, requested):
             try:
                 get_value_ext = prop.get_value_ext
             except AttributeError:
-                prop.get_value(href, resource, ret)
+                prop.get_value(href, resource, ret, environ)
             else:
-                get_value_ext(href, resource, ret, requested)
+                get_value_ext(href, resource, ret, environ, requested)
         except KeyError:
             statuscode = '404 Not Found'
         else:
@@ -904,44 +954,48 @@ def get_property_from_element(href, resource, properties, requested):
     return PropStatus(statuscode, responsedescription, ret)
 
 
-def get_properties(href, resource, properties, requested):
+def get_properties(href, resource, properties, environ, requested):
     """Get a set of properties.
 
     :param href: Resource Href
     :param resource: Resource object
     :param properties: Dictionary of properties
     :param requested: XML {DAV:}prop element with properties to look up
+    :param environ: WSGI environ dict
     :return: Iterator over PropStatus items
     """
     for propreq in list(requested):
-        yield get_property_from_element(href, resource, properties, propreq)
+        yield get_property_from_element(
+            href, resource, properties, environ, propreq)
 
 
-def get_property_names(href, resource, properties, requested):
+def get_property_names(href, resource, properties, environ, requested):
     """Get a set of property names.
 
     :param href: Resource Href
     :param resource: Resource object
     :param properties: Dictionary of properties
+    :param environ: WSGI environ dict
     :param requested: XML {DAV:}prop element with properties to look up
     :return: Iterator over PropStatus items
     """
     for name, prop in properties.items():
-        if prop.is_set(href, resource):
+        if prop.is_set(href, resource, environ):
             yield PropStatus('200 OK', None, ET.Element(name))
 
 
-def get_all_properties(href, resource, properties):
+def get_all_properties(href, resource, properties, environ):
     """Get all properties.
 
     :param href: Resource Href
     :param resource: Resource object
     :param properties: Dictionary of properties
     :param requested: XML {DAV:}prop element with properties to look up
+    :param environ: WSGI environ dict
     :return: Iterator over PropStatus items
     """
     for name in properties:
-        ps = get_property_from_name(href, resource, properties, name)
+        ps = get_property_from_name(href, resource, properties, name, environ)
         if ps.statuscode == '200 OK':
             yield ps
 
@@ -1001,8 +1055,12 @@ class Reporter(object):
         :param resource: Resource to check for
         :return: boolean indicating whether this reporter is available
         """
-        return (self.resource_type is None or
-                self.resource_type in resource.resource_types)
+        if self.resource_type is None:
+            return True
+        if isinstance(self.resource_type, tuple):
+            return any(rs in resource.resource_types
+                       for rs in self.resource_type)
+        return self.resource_type in resource.resource_types
 
     def report(self, environ, start_response, request_body, resources_by_hrefs,
                properties, href, resource, depth):
@@ -1044,7 +1102,7 @@ class ExpandPropertyReporter(Reporter):
     name = '{DAV:}expand-property'
 
     def _populate(self, prop_list, resources_by_hrefs, properties, href,
-                  resource):
+                  resource, environ):
         """Expand properties for a resource.
 
         :param prop_list: DAV:property elements to retrieve and expand
@@ -1052,6 +1110,7 @@ class ExpandPropertyReporter(Reporter):
         :param properties: Available properties
         :param href: href for current resource
         :param resource: current resource
+        :param environ: WSGI environ dict
         :return: Status object
         """
         ret = []
@@ -1059,7 +1118,7 @@ class ExpandPropertyReporter(Reporter):
             prop_name = prop.get('name')
             # FIXME: Resolve prop_name on resource
             propstat = get_property_from_name(
-                href, resource, properties, prop_name)
+                href, resource, properties, prop_name, environ)
             new_prop = ET.Element(propstat.prop.tag)
             child_hrefs = [
                 read_href_element(prop_child)
@@ -1078,7 +1137,8 @@ class ExpandPropertyReporter(Reporter):
                         new_prop.append(prop_child)
                     else:
                         response = self._populate(
-                            prop, properties, child_href, child_resource)
+                            prop, properties, child_href, child_resource,
+                            environ)
                         new_prop.append(response.aselement())
             propstat = PropStatus(propstat.statuscode,
                                   propstat.responsedescription, prop=new_prop)
@@ -1089,7 +1149,7 @@ class ExpandPropertyReporter(Reporter):
     def report(self, environ, request_body, resources_by_hrefs, properties,
                href, resource, depth):
         return self._populate(request_body, resources_by_hrefs, properties,
-                              href, resource)
+                              href, resource, environ)
 
 
 class SupportedLockProperty(Property):
@@ -1102,7 +1162,7 @@ class SupportedLockProperty(Property):
     resource_type = None
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         for (lockscope, locktype) in resource.get_supported_locks():
             entry = ET.SubElement(el, '{DAV:}lockentry')
             scope_el = ET.SubElement(entry, '{DAV:}lockscope')
@@ -1121,7 +1181,7 @@ class LockDiscoveryProperty(Property):
     resource_type = None
     live = True
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         for activelock in resource.get_active_locks():
             entry = ET.SubElement(el, '{DAV:}activelock')
             type_el = ET.SubElement(entry, '{DAV:}locktype')
@@ -1150,7 +1210,7 @@ class CommentProperty(Property):
     live = False
     in_allprops = False
 
-    def get_value(self, href, resource, el):
+    def get_value(self, href, resource, el, environ):
         el.text = resource.get_comment()
 
     def set_value(self, href, resource, el):
@@ -1191,6 +1251,8 @@ def _get_resources_by_hrefs(backend, environ, hrefs):
 
 def _send_xml_response(start_response, status, et, out_encoding):
     body_type = 'text/xml; charset="%s"' % out_encoding
+    if os.environ.get('XANDIKOS_DUMP_DAV_XML'):
+        print("OUT: " + ET.tostring(et).decode('utf-8'))
     body = ET.tostringlist(et, encoding=out_encoding)
     start_response(status, [
         ('Content-Type', body_type),
@@ -1303,6 +1365,8 @@ def _readXmlBody(environ, expected_tag=None):
         if base_content_type not in ('text/xml', 'application/xml'):
             raise UnsupportedMediaType(content_type)
     body = b''.join(_readBody(environ))
+    if os.environ.get('XANDIKOS_DUMP_DAV_XML'):
+        print("IN: " + body.decode('utf-8'))
     try:
         et = xmlparse(body)
     except ET.ParseError:
@@ -1490,13 +1554,13 @@ class PropfindMethod(Method):
             propstat = []
             if requested is None or requested.tag == '{DAV:}allprop':
                 propstat = get_all_properties(
-                    href, resource, app.properties)
+                    href, resource, app.properties, environ)
             elif requested.tag == '{DAV:}prop':
                 propstat = get_properties(
-                    href, resource, app.properties, requested)
+                    href, resource, app.properties, environ, requested)
             elif requested.tag == '{DAV:}propname':
                 propstat = get_property_names(
-                    href, resource, app.properties, requested)
+                    href, resource, app.properties, environ, requested)
             else:
                 raise BadRequestError(
                     'Expected prop/allprop/propname tag, got ' + requested.tag)
@@ -1694,8 +1758,9 @@ class WebDAVApp(object):
 
     def _get_dav_features(self, resource):
         # TODO(jelmer): Support access-control
-        return ['1', '2', '3', 'calendar-access', 'addressbook',
-                'extended-mkcol']
+        return ['1', '2', '3', 'calendar-access', 'calendar-auto-scheduling',
+                'addressbook', 'extended-mkcol', 'add-member',
+                'sync-collection', 'quota']
 
     def _get_allowed_methods(self, environ):
         """List of supported methods on this endpoint."""
@@ -1709,6 +1774,9 @@ class WebDAVApp(object):
         if environ.get('HTTP_EXPECT', '') != '':
             start_response('417 Expectation Failed', [])
             return []
+        if 'SCRIPT_NAME' not in environ:
+            logging.debug('SCRIPT_NAME not set; assuming "".')
+            environ['SCRIPT_NAME'] = ''
         method = environ['REQUEST_METHOD']
         try:
             do = self.methods[method]
